/*
 * USB Nomad Jukebox filesystem for Linux.
 *
 * Copyright (C) 2001 Jrg Prante <joergprante@gmx.de>
 *
 * based on ftpfs by Florin Malita <fmalita@yahoo.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/smp_lock.h>
#include <linux/fs.h>
#include <linux/usb.h>
#include <linux/slab.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include "njbfs.h"
#include "njbfs_proc.h"

#define dir_cache info->cache

void njbfs_cache_init(struct njbfs_sb_info *info)
{
	memset(&dir_cache, 0, sizeof(struct njbfs_dir_cache));
}

unsigned long njbfs_cache_hash(char *name)
{
	unsigned long hash = 0;
	int i;

	for (i = 0; i < strlen(name); i++)
		hash += name[i];
	return hash % NJBFS_CACHE_HASH;
}

int njbfs_cache_deldir(struct njbfs_hashlist_node *n)
{
	struct njbfs_dirlist_node *p, *q;

	if (!n)
		return -1;
	for (p = n->directory.head; p != NULL; p = q) {
		q = p->next;
		if (p->entry.name != NULL) {
			kfree(p->entry.name);
			p->entry.name = NULL;
		}
		kfree(p);
	}

	kfree(n->directory.name);
	if (n->prev != NULL)
		n->prev->next = n->next;
	if (n->next != NULL)
		n->next->prev = n->prev;
	return 0;
}

int njbfs_cache_infront(unsigned long hsh, struct njbfs_hashlist_node *n,
			struct njbfs_sb_info *info)
{
	unsigned long h;

	if (n->prev) {
		h = njbfs_cache_hash(n->directory.name);

		n->prev->next = n->next;

		if (n->next)
			n->next->prev = n->prev;

		if (dir_cache.hash[h])
			dir_cache.hash[h]->prev = n;
		n->prev = NULL;
		n->next = dir_cache.hash[h];
		dir_cache.hash[h] = n;
	}
	return 0;
}


int njbfs_cache_shrink(unsigned long hsh, struct njbfs_sb_info *info)
{
	struct njbfs_hashlist_node *p;

	if (dir_cache.len[hsh] == 0)
		return -1;
	for (p = dir_cache.hash[hsh]; p->next != NULL; p = p->next);
	njbfs_cache_deldir(p);
	dir_cache.len[hsh]--;

	return 0;
}

int njbfs_cache_add(struct njbfs_sb_info *info, char *name,
		    struct njbfs_directory **dir)
{
	unsigned long hsh;
	struct njbfs_hashlist_node *p;
	int res;

	hsh = njbfs_cache_hash(name);

	while (dir_cache.len[hsh] >= NJBFS_CACHE_LEN)
		njbfs_cache_shrink(hsh, info);
	p =
	    (struct njbfs_hashlist_node *)
	    kmalloc(sizeof(struct njbfs_hashlist_node), GFP_KERNEL);
	if (!p) {
		res = -1;
		err("cache_add: kmalloc error!");
		goto out;
	}

	memset(p, 0, sizeof(struct njbfs_hashlist_node));
	p->directory.valid = 1;
	p->directory.time = CURRENT_TIME;
	p->directory.name = (char *) kmalloc(strlen(name) + 1, GFP_KERNEL);
	if (!p->directory.name) {
		err("kmalloc error!");
		res = -1;
		goto out1;
	}
	strcpy(p->directory.name, name);

	info("cache_add: loading directory '%s'....", name);
	if ((res = njbfs_loaddir(info, name, &p->directory)) < 0) {
		err("cache_add: couldn't load directory");
		goto out2;
	}
	p->prev = NULL;
	p->next = dir_cache.hash[hsh];
	if (p->next)
		p->next->prev = p;
	dir_cache.hash[hsh] = p;

	dir_cache.len[hsh]++;
	*dir = &p->directory;
	return 0;
      out2:
	kfree(p->directory.name);
      out1:
	kfree(p);
      out:
	return res;
}

int njbfs_cache_get(struct njbfs_sb_info *info, char *name,
		    struct njbfs_directory **dir)
{
	struct njbfs_hashlist_node *p;
	unsigned long hsh;

	hsh = njbfs_cache_hash(name);
	for (p = dir_cache.hash[hsh]; p != NULL; p = p->next)
		if ((strcmp(name, p->directory.name) == 0)
		    && (p->directory.valid)) {
			if (CURRENT_TIME - p->directory.time >
			    NJBFS_CACHE_TTL) {
				/* entry too old */
				//return  njbfs_cache_add(info, name, dir);
			}
			njbfs_cache_infront(hsh, p, info);
			*dir = &p->directory;
			return 0;
		}

	return njbfs_cache_add(info, name, dir);
}

int njbfs_cache_empty(struct njbfs_sb_info *info)
{
	int i, j;
	struct njbfs_hashlist_node *p;

	for (i = 0; i < NJBFS_CACHE_HASH; i++)
		for (j = 0; j < dir_cache.len[i]; j++) {
			p = dir_cache.hash[i];
			dir_cache.hash[i] = p->next;
			njbfs_cache_deldir(p);
		}
	return 0;
}

void njbfs_cache_invalidate(struct dentry *dentry)
{
	struct super_block *sb = dentry->d_inode->i_sb;
	struct njbfs_sb_info *info =
	    (struct njbfs_sb_info *) sb->u.generic_sbp;
	char buf[NJBFS_MAXPATHLEN];
	struct njbfs_hashlist_node *p;
	unsigned long hsh;

	njbfs_get_name(dentry, buf);
	hsh = njbfs_cache_hash(buf);

	for (p = dir_cache.hash[hsh]; p != NULL; p = p->next)
		if (strcmp(buf, p->directory.name) == 0) {
			p->directory.valid = 0;
			return;
		}
}
