/*
 * USB Nomad Jukebox filesystem for Linux.
 *
 * Copyright (C) 2001 Jrg Prante <joergprante@gmx.de>
 *
 * based on ftpfs by Florin Malita <fmalita@yahoo.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/smp_lock.h>
#include <linux/fs.h>
#include <linux/usb.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include "njbfs.h"
#include "njbfs_proc.h"

static int njbfs_lookup_validate(struct dentry *, int);
static int njbfs_hash_dentry(struct dentry *, struct qstr *);
static int njbfs_compare_dentry(struct dentry *, struct qstr *,
				struct qstr *);
static int njbfs_delete_dentry(struct dentry *);

static struct dentry_operations njbfs_dentry_operations = {
	d_revalidate:njbfs_lookup_validate,
	d_hash:njbfs_hash_dentry,
	d_compare:njbfs_compare_dentry,
	d_delete:njbfs_delete_dentry,
};

static int njbfs_readdir(struct file *, void *, filldir_t);
static int njbfs_dir_open(struct inode *, struct file *);

static struct dentry *njbfs_lookup(struct inode *, struct dentry *);
static int njbfs_mkdir(struct inode *, struct dentry *, int);
static int njbfs_create(struct inode *, struct dentry *, int);
static int njbfs_rmdir(struct inode *, struct dentry *);
static int njbfs_rename(struct inode *, struct dentry *, struct inode *,
			struct dentry *);
static int njbfs_unlink(struct inode *, struct dentry *);

struct file_operations njbfs_dir_operations = {
	read:generic_read_dir,
	readdir:njbfs_readdir,
	open:njbfs_dir_open,
};

struct inode_operations njbfs_dir_inode_operations = {
	create:njbfs_create,
	lookup:njbfs_lookup,
	unlink:njbfs_unlink,
	mkdir:njbfs_mkdir,
	rmdir:njbfs_rmdir,
	rename:njbfs_rename,
};


static int njbfs_lookup_validate(struct dentry *dentry, int flags)
{
	struct inode *inode = dentry->d_inode;
	int valid = 0;

	if (inode) {
		lock_kernel();
		if (is_bad_inode(inode))
			valid = 0;
		unlock_kernel();
	}

	return valid;
}

static int njbfs_hash_dentry(struct dentry *dentry, struct qstr *this)
{
	unsigned long hash;
	int i;

	hash = init_name_hash();
	for (i = 0; i < this->len; i++)
		hash = partial_name_hash(this->name[i], hash);
	this->hash = end_name_hash(hash);
	return 0;
}

static int
njbfs_compare_dentry(struct dentry *dentry, struct qstr *a, struct qstr *b)
{
	int i, res = 1;

	if (a->len != b->len)
		goto out;
	for (i = 0; i < a->len; i++)
		if (a->name[i] != b->name[i])
			goto out;
	res = 0;
      out:
	return res;
}

static int njbfs_delete_dentry(struct dentry *dentry)
{
	if (dentry->d_inode) {
		if (is_bad_inode(dentry->d_inode))
			return 1;
	}
	return 0;
}

static int njbfs_readdir(struct file *f, void *dirent, filldir_t filldir)
{
	struct dentry *dentry = f->f_dentry;
	struct inode *inode = dentry->d_inode;
	struct super_block *sb = inode->i_sb;
	struct njbfs_sb_info *info =
	    (struct njbfs_sb_info *) sb->u.generic_sbp;
	char buf[NJBFS_MAXPATHLEN];
	int result, pos, res;

	struct njbfs_directory *dir;
	struct njbfs_dirlist_node *file;

	if (!info->mnt.mount_point[0]) {
		njbfs_get_name(f->f_vfsmnt->mnt_mountpoint,
			       info->mnt.mount_point);
	}

	dbg("reading %s, f_pos=%d", dentry->d_name.name, (int) f->f_pos);
	if (njbfs_get_name(dentry, buf) < 0) {
		err(" njbfs_get_name failed");
		return -1;
	}

	result = 0;
	switch ((unsigned int) f->f_pos) {
	case 0:
		if (filldir(dirent, ".", 1, 0, inode->i_ino, DT_DIR) < 0)
			goto out;
		f->f_pos = 1;
	case 1:
		if (filldir
		    (dirent, "..", 2, 1, dentry->d_parent->d_inode->i_ino,
		     DT_DIR) < 0)
			goto out;
		f->f_pos = 2;
	default:
		njbfs_lock(info);
		res = njbfs_cache_get(info, buf, &dir);
		njbfs_unlock(info);
		if (res < 0) {
			info(" njbfs_cache_get failed");
			switch (res) {
			case -ERESTARTSYS:
				res = -EINTR;
				njbfs_disconnect(info);
				break;
			}
			return res;
		}

		pos = 2;
		for (file = dir->head; file != NULL; file = file->next) {
			if ((pos == f->f_pos)) {
				struct qstr qname;
				unsigned long ino;

				qname.name = file->entry.name;
				qname.len = strlen(qname.name);

				ino = find_inode_number(dentry, &qname);
				if (!ino)
					ino = iunique(dentry->d_sb, 2);
				if (filldir
				    (dirent, qname.name, qname.len,
				     f->f_pos, ino, DT_UNKNOWN) >= 0)
					f->f_pos++;

			}
			pos++;
		}

	}
      out:
	return result;
}

static int njbfs_dir_open(struct inode *inode, struct file *f)
{
	return 0;
}


static struct dentry *njbfs_lookup(struct inode *dir,
				   struct dentry *dentry)
{
	struct super_block *sb = dir->i_sb;
	struct njbfs_sb_info *info =
	    (struct njbfs_sb_info *) sb->u.generic_sbp;
	struct njbfs_fattr fattr;
	struct inode *inode;
	int res;

	dbg("dname:%s", dentry->d_name.name);
	if ((res = njbfs_get_attr(dentry, &fattr,
				  (struct njbfs_sb_info *) dir->i_sb->u.
				  generic_sbp)) < 0) {
		inode = NULL;
		dentry->d_op = &njbfs_dentry_operations;
		d_add(dentry, inode);
		switch (res) {
		case -ERESTARTSYS:
			res = -EINTR;
			njbfs_disconnect(info);
			return ERR_PTR(res);
		}
		return NULL;
	}

	fattr.f_ino = iunique(dentry->d_sb, 2);
	inode = njbfs_iget(dir->i_sb, &fattr);
	if (inode) {
		dentry->d_op = &njbfs_dentry_operations;
		d_add(dentry, inode);
	}

	return NULL;
}

static int njbfs_instantiate(struct dentry *dentry)
{
	struct njbfs_sb_info *info =
	    (struct njbfs_sb_info *) dentry->d_sb->u.generic_sbp;
	struct njbfs_fattr fattr;
	struct inode *inode;

	if (njbfs_get_attr(dentry, &fattr, info) < 0) {
		info("njbfs_get_attr failed");
		return -1;
	}

	fattr.f_ino = iunique(dentry->d_sb, 2);
	inode = njbfs_iget(dentry->d_sb, &fattr);
	if (!inode)
		return -EACCES;

	d_instantiate(dentry, inode);

	return 0;
}

static int njbfs_mkdir(struct inode *dir, struct dentry *dentry, int mode)
{
	struct njbfs_sb_info *info =
	    (struct njbfs_sb_info *) dentry->d_sb->u.generic_sbp;
	char buf[NJBFS_MAXPATHLEN];
	int res;

	njbfs_get_name(dentry, buf);
	if ((res = njbfs_proc_mkdir(info, buf)) < 0) {
		err("mkdir failed");
		return res;
	}
	njbfs_cache_invalidate(dentry->d_parent);
	return njbfs_instantiate(dentry);
}

static int njbfs_create(struct inode *dir, struct dentry *dentry, int mode)
{
	struct njbfs_sb_info *info =
	    (struct njbfs_sb_info *) dentry->d_sb->u.generic_sbp;
	char buf[NJBFS_MAXPATHLEN];
	int res;

	njbfs_get_name(dentry, buf);
	if ((res = njbfs_proc_create(info, buf)) < 0) {
		err("create failed");
		return res;
	}

	njbfs_cache_invalidate(dentry->d_parent);
	return njbfs_instantiate(dentry);
}

static int njbfs_rmdir(struct inode *dir, struct dentry *dentry)
{
	struct njbfs_sb_info *info =
	    (struct njbfs_sb_info *) dentry->d_sb->u.generic_sbp;
	char buf[NJBFS_MAXPATHLEN];

	if (!d_unhashed(dentry))
		return -EBUSY;

	njbfs_get_name(dentry, buf);

	njbfs_cache_invalidate(dentry->d_parent);

	return njbfs_proc_rmdir(info, buf);
}

static int
njbfs_rename(struct inode *old_dir, struct dentry *old_dentry,
	     struct inode *new_dir, struct dentry *new_dentry)
{
	struct njbfs_sb_info *info =
	    (struct njbfs_sb_info *) old_dentry->d_sb->u.generic_sbp;
	int res;

	if (new_dentry->d_inode) {
		d_delete(new_dentry);
	}

	if (
	    (res =
	     njbfs_proc_rename(info, old_dir, old_dentry, new_dir,
			       new_dentry)) < 0) return res;

	njbfs_cache_invalidate(old_dentry->d_parent);
	njbfs_cache_invalidate(new_dentry->d_parent);

	return 0;
}

static int njbfs_unlink(struct inode *dir, struct dentry *dentry)
{
	struct njbfs_sb_info *info =
	    (struct njbfs_sb_info *) dentry->d_sb->u.generic_sbp;
	char buf[NJBFS_MAXPATHLEN];

	njbfs_get_name(dentry, buf);
	njbfs_cache_invalidate(dentry->d_parent);
	return njbfs_proc_unlink(info, buf);
}
