/*
 * USB Nomad Jukebox filesystem for Linux.
 *
 * Copyright (C) 2001 Jrg Prante <joergprante@gmx.de>
 *
 * based on ftpfs by Florin Malita <fmalita@yahoo.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <linux/version.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pagemap.h>
#include <linux/smp_lock.h>
#include <linux/signal.h>
#include <linux/usb.h>

#include <asm/uaccess.h>
#include <asm/system.h>

#include "njbfs.h"
#include "njbfs_proc.h"

extern int njbfs_open(struct inode *inode, struct file *file);
extern int njbfs_close(struct inode *inode, struct file *file);

static int njbfs_file_readpage(struct file *f, struct page *p)
{
	struct dentry *dentry = f->f_dentry;
	char *buffer;
	unsigned long offset, count;
	int result;

	get_page(p);

	buffer = page_address(p);
	offset = p->index << PAGE_CACHE_SHIFT;
	count = PAGE_SIZE;

	do {
		result = njbfs_read(dentry, offset, count, buffer);
		if (result < 0) {
			err("read error");
			goto io_error;
		}
		count -= result;
		offset += result;
		buffer += result;
		dentry->d_inode->i_atime = CURRENT_TIME;
		if (!result)
			break;
	} while (count);

	memset(buffer, 0, count);
	flush_dcache_page(p);
	SetPageUptodate(p);
	result = 0;
      io_error:
	UnlockPage(p);
	put_page(p);
	switch (result) {
	case -ERESTARTSYS:
		if (sigismember(&current->pending.signal, SIGPIPE)) {
			sigdelset(&current->pending.signal, SIGPIPE);
			current->sigpending--;
		}

		result = -EINTR;
		break;
	}
	return result;
}

static int njbfs_file_writepage(struct page *p)
{
	dbg("This shouldn't happen");
	return -EFAULT;
}

static int
njbfs_file_preparewrite(struct file *f, struct page *p, unsigned offset,
			unsigned to)
{
	kmap(p);
	return 0;
}

static int
njbfs_file_commitwrite(struct file *f, struct page *p, unsigned offset,
		       unsigned to)
{
	struct dentry *dentry = f->f_dentry;
	char *buffer = page_address(p) + offset;
	int written = 0, result;
	unsigned count = to - offset;

	offset += p->index << PAGE_CACHE_SHIFT;

	lock_kernel();

	do {
		result = njbfs_write(dentry, offset, count, buffer);
		if (result < 0) {
			err("write error");
			goto error;
		}

		count -= result;
		offset += result;
		buffer += result;
		written += result;
		dentry->d_inode->i_mtime = dentry->d_inode->i_atime =
		    CURRENT_TIME;
		if (!result)
			break;

	} while (count);

	memset(buffer, 0, count);
	result = 0;
      error:
	unlock_kernel();
	kunmap(p);

	switch (result) {
	case -ERESTARTSYS:
		if (sigismember(&current->pending.signal, SIGPIPE)) {
			sigdelset(&current->pending.signal, SIGPIPE);
			current->sigpending--;
		}

		result = -EINTR;
		break;
	}

	return (result == 0) ? written : result;
}

static int njbfs_file_permission(struct inode *inode, int mask)
{
	int mode = inode->i_mode;

	mode >>= 6;
	if ((mode & 7 & mask) != mask)
		return -EACCES;
	return 0;
}

static int njbfs_file_open(struct inode *inode, struct file *f)
{
	return njbfs_open(inode, f);
}

static int njbfs_file_release(struct inode *inode, struct file *f)
{
	return njbfs_close(inode, f);
}

struct file_operations njbfs_file_operations = {
	read:generic_file_read,
	write:generic_file_write,
	mmap:generic_file_mmap,
	open:njbfs_file_open,
	release:njbfs_file_release,
};

struct inode_operations njbfs_file_inode_operations = {
	permission:njbfs_file_permission,
};

struct address_space_operations njbfs_file_aops = {
	readpage:njbfs_file_readpage,
	writepage:njbfs_file_writepage,
	prepare_write:njbfs_file_preparewrite,
	commit_write:njbfs_file_commitwrite
};
