/*
 * USB Nomad Jukebox filesystem for Linux.
 *
 * Copyright (C) 2001 Jrg Prante <joergprante@gmx.de>
 *
 * based on ftpfs by Florin Malita <fmalita@yahoo.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/smp_lock.h>
#include <linux/fs.h>
#include <linux/usb.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include "njbfs.h"
#include "njbfs_proc.h"
#include "nomad_usb.h"

MODULE_AUTHOR("Jrg Prante <joergprante@gmx.de>");
MODULE_DESCRIPTION("USB Nomad Jukebox File System for Linux");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

/* from nomad.o */
extern int usb_nomad_init(void);
extern void usb_nomad_cleanup(void);
extern struct nomad_usb_data nomad_instance;

#if defined(CONFIG_PROC_FS)
extern int proc_njb_create(struct nomad_usb_data *njb);
extern void proc_njb_destroy(void);
#endif


extern struct inode_operations njbfs_dir_inode_operations;
extern struct file_operations njbfs_dir_operations;
extern struct inode_operations njbfs_file_inode_operations;
extern struct file_operations njbfs_file_operations;
extern struct address_space_operations njbfs_file_aops;
extern struct inode_operations njbfs_symlink_inode_operations;

static void njbfs_delete_inode(struct inode *);
static void njbfs_put_super(struct super_block *);
static int njbfs_statfs(struct super_block *, struct statfs *);
static void njbfs_set_inode_attr(struct inode *, struct njbfs_fattr *);

static struct super_operations njbfs_sops = {
	put_inode:force_delete,
	delete_inode:njbfs_delete_inode,
	put_super:njbfs_put_super,
	statfs:njbfs_statfs,
};


struct inode *njbfs_iget(struct super_block *sb, struct njbfs_fattr *fattr)
{
	struct inode *res;

	res = new_inode(sb);
	if (!res)
		return NULL;
	res->i_ino = fattr->f_ino;
	njbfs_set_inode_attr(res, fattr);

	if (S_ISDIR(res->i_mode)) {
		res->i_op = &njbfs_dir_inode_operations;
		res->i_fop = &njbfs_dir_operations;;
	} else if (S_ISLNK(res->i_mode)) {
		res->i_op = &njbfs_symlink_inode_operations;
	} else {
		res->i_op = &njbfs_file_inode_operations;
		res->i_fop = &njbfs_file_operations;
		res->i_data.a_ops = &njbfs_file_aops;
	}

	insert_inode_hash(res);
	return res;
}

static void
njbfs_set_inode_attr(struct inode *inode, struct njbfs_fattr *fattr)
{
	inode->i_mode = fattr->f_mode;
	inode->i_nlink = fattr->f_nlink;
	inode->i_uid = fattr->f_uid;
	inode->i_gid = fattr->f_gid;
	inode->i_rdev = fattr->f_rdev;
	inode->i_ctime = fattr->f_ctime;
	inode->i_atime = fattr->f_atime;
	inode->i_mtime = fattr->f_mtime;
	inode->i_blksize = fattr->f_blksize;
	inode->i_blocks = fattr->f_blocks;
	inode->i_size = fattr->f_size;
}


static void njbfs_delete_inode(struct inode *inode)
{
	lock_kernel();
	clear_inode(inode);
	unlock_kernel();
}

static void njbfs_put_super(struct super_block *sb)
{
	struct njbfs_sb_info *info =
	    (struct njbfs_sb_info *) sb->u.generic_sbp;

	njbfs_disconnect(info);
	njbfs_cache_empty(info);
	kfree(info);
	info("super block discarded");
}

static int njbfs_statfs(struct super_block *sb, struct statfs *attr)
{
	attr->f_type = NJBFS_SUPER_MAGIC;
	attr->f_bsize = 1024;
	attr->f_blocks = 0;
	attr->f_namelen = NJBFS_MAXPATHLEN;
	attr->f_files = -1;
	attr->f_bavail = -1;

	return 0;
}

struct super_block *njbfs_read_super(struct super_block *sb, void *opts,
				     int silent)
{
	struct njbfs_sb_info *info;
	struct njbfs_fattr root;
	struct inode *root_inode;

	info =
	    (struct njbfs_sb_info *) kmalloc(sizeof(struct njbfs_sb_info),
					     GFP_KERNEL);
	if (!info) {
		info("not enough kmem to allocate info");
		goto out;
	}
	memset(info, 0, sizeof(struct njbfs_sb_info));

	sb->u.generic_sbp = info;
	sb->s_blocksize = 1024;
	sb->s_blocksize_bits = 10;
	sb->s_magic = NJBFS_SUPER_MAGIC;
	sb->s_op = &njbfs_sops;
	sb->s_flags = 0;

	info->mnt.active = 0;
	info->mnt.file_mode = (S_IRUSR | S_IFREG);
	info->mnt.dir_mode = (S_IRUSR | S_IXUSR | S_IFDIR);
	info->mnt.uid = current->uid;
	info->mnt.gid = current->gid;

	info->njb = &nomad_instance;

	info->virtual_mode = 0;
	strcpy(info->virtual_dir, "");

	init_MUTEX(&info->sem);

	info->mnt.mount_point[0] = 0;

	njbfs_cache_init(info);

	if (njbfs_parse_options(info, opts) < 0) {
		dbg("invalid options");
		goto out_no_opts;
	}

	if (njbfs_connect(info) < 0) {
		info("could not connect");
		goto out_no_opts;
	}

	njbfs_init_root_dirent(info, &root);
	root_inode = njbfs_iget(sb, &root);
	if (!root_inode)
		goto out_no_root;

	sb->s_root = d_alloc_root(root_inode);
	if (!sb->s_root)
		goto out_no_root;

	dbg("mount succeded");
	return sb;

      out_no_root:
	iput(root_inode);
      out_no_opts:
	kfree(info);
      out:
	dbg("mount failed");
	return NULL;
}

static DECLARE_FSTYPE(njbfs_fs_type, "njbfs", njbfs_read_super, 0);



static int init_njb_module(void)
{
	int ret;

	ret = usb_nomad_init();
	if (ret < 0)
		return ret;

#if defined(CONFIG_PROC_FS)
	ret = proc_njb_create(&nomad_instance);
	if (ret < 0)
		return ret;
#endif

	return register_filesystem(&njbfs_fs_type);
}

static void exit_njb_module(void)
{
	unregister_filesystem(&njbfs_fs_type);
#if defined(CONFIG_PROC_FS)
	proc_njb_destroy();
#endif
	usb_nomad_cleanup();

	return;
}


EXPORT_NO_SYMBOLS;

module_init(init_njb_module);
module_exit(exit_njb_module);
