/*
 * USB Nomad Jukebox filesystem for Linux.
 *
 * Copyright (C) 2001 Jrg Prante <joergprante@gmx.de>
 *
 * based on ftpfs by Florin Malita <fmalita@yahoo.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/types.h>

#include "njbfs.h"

static playlist_t *playlist_new(void);
static void playlist_destroy(playlist_t * pl);
static void playlist_addtrack(playlist_t * pl, playlist_track_t * track,
			      unsigned int pos);
#define NJB_PL_END	0
#define NJB_PL_START	1
static void playlist_reset_gettrack(playlist_t * pl);
static playlist_track_t *playlist_gettrack(playlist_t * pl);
static int playlist_set_name(playlist_t * pl, const char *name);
static void playlist_deltrack(playlist_t * pl, unsigned int pos);
static playlist_t *playlist_unpack(void *data, size_t nbytes);
static playlist_track_t *playlist_track_new(u_int32_t trackid);
static void playlist_track_destroy(playlist_track_t * pl);


/**
 *   New internal playlist
 */

static playlist_t *playlist_new(void)
{
	playlist_t *playlist;

	if (
	    (playlist =
	     (playlist_t *) kmalloc(sizeof(playlist_t),
				    GFP_KERNEL)) == NULL) {
		return NULL;
	}
	memset(playlist, 0, sizeof(playlist_t));

	return playlist;
}

/**
 *  Unpack a playlist received from NJB
 */
static playlist_t *playlist_unpack(void *data, size_t nbytes)
{
	unsigned char *dp = (unsigned char *) data;
	size_t index;
	playlist_t *playlist;
	u_int32_t ntracks, i;
	u_int16_t lname;

	/* new internal playlist */
	playlist = playlist_new();
	if (playlist == NULL)
		return NULL;


	memcpy(&playlist->plid, dp, 4);
	memcpy(&lname, &dp[4], 2);
	playlist->name = (char *) kmalloc(lname, GFP_KERNEL);
	if (playlist->name == NULL) {
		playlist_destroy(playlist);
		return NULL;
	}

	memcpy(playlist->name, &dp[6], lname);
	index = lname + 12;
	memcpy(&ntracks, &dp[index], 4);
	index += 4;

	for (i = 0; i < ntracks; i++) {
		u_int32_t trackid;
		playlist_track_t *track;
		index += 4;
		memcpy(&trackid, &dp[index], 4);
		index += 4;
		if (index > nbytes) {
			playlist_destroy(playlist);
			return NULL;
		}
		track = playlist_track_new(trackid);
		if (track == NULL) {
			playlist_destroy(playlist);
			return NULL;
		}

		playlist_addtrack(playlist, track, NJB_PL_END);
	}

	playlist->_state = NJB_PL_UNCHANGED;
	return playlist;
}

/**
 *   Add track to playlist
 */
static void playlist_addtrack(playlist_t * pl, playlist_track_t * track,
			      unsigned int pos)
{
	if (pl->_state != NJB_PL_NEW)
		pl->_state = NJB_PL_CHTRACKS;

	if (pos > pl->ntracks)
		pos = NJB_PL_END;

	if (pos == NJB_PL_END) {
		if (pl->first == NULL) {
			pl->first = pl->cur = pl->last = track;
			track->next = track->prev = NULL;
		} else {
			track->prev = pl->last;
			track->next = NULL;
			pl->last->next = track;
			pl->last = track;
		}
	} else if (pos == NJB_PL_START) {
		if (pl->first == NULL) {
			pl->first = pl->cur = pl->last = track;
			track->next = track->prev = NULL;
		} else {
			track->prev = NULL;
			track->next = pl->first;
			pl->first->prev = track;
			pl->first = track;
		}
	} else {
		int i = 1;
		playlist_track_t *cur;

		playlist_reset_gettrack(pl);
		while ((cur = playlist_gettrack(pl)) != NULL) {
			if (i == pos) {
				cur->prev->next = track;
				track->prev = cur->prev;
				track->next = cur;
				cur->prev = track;

				pl->ntracks++;

				return;
			}

			i++;
		}
	}

	pl->ntracks++;
}

/**
 *  Delete track from playlist
 *
 *  @param pl the playlist
 *  @param pos the position of the track in the playlist
 */
static void playlist_deltrack(playlist_t * pl, unsigned int pos)
{
	playlist_track_t *track;

	if (pos > pl->ntracks)
		pos = NJB_PL_END;

	pl->_state = NJB_PL_CHTRACKS;

	if (pos == NJB_PL_START) {
		track = pl->first;
		pl->first = pl->first->next;
		pl->first->prev = NULL;
	} else if (pos == NJB_PL_END) {
		track = pl->last;
		pl->last = pl->last->prev;
		pl->last->next = NULL;
	} else {
		int i = 1;

		playlist_reset_gettrack(pl);
		while ((track = playlist_gettrack(pl)) != NULL) {
			if (i == pos) {
				track->prev->next = track->next;
				track->next->prev = track->prev;

				playlist_track_destroy(track);
				pl->ntracks--;

				return;
			}
			i++;
		}
	}

	playlist_track_destroy(track);
	pl->ntracks--;
}

/**
 *  Destroy playlist
 */
static void playlist_destroy(playlist_t * pl)
{
	playlist_track_t *pltrack;

	pl->cur = pl->first;
	while (pl->cur != NULL) {
		pltrack = pl->cur;
		pl->cur = pl->cur->next;
		playlist_track_destroy(pltrack);
	}

	if (pl->name != NULL)
		kfree(pl->name);

	kfree(pl);
}

/**
 *  Reset playlist to get tracks
 */
static void playlist_reset_gettrack(playlist_t * pl)
{
	pl->cur = pl->first;
}

/**
 *  Get current track from playlist
 */
static playlist_track_t *playlist_gettrack(playlist_t * pl)
{
	playlist_track_t *track;

	if (pl->cur == NULL)
		return NULL;

	track = pl->cur;
	pl->cur = pl->cur->next;

	return track;
}


/**
 *  Set name of playlist
 */
static int playlist_set_name(playlist_t * pl, const char *name)
{

	char *newname;

	newname = (char *) kmalloc(strlen(name), GFP_KERNEL);
	if (newname == NULL) {
		return -1;
	}
	memcpy(newname, name, strlen(name));

	if (pl->name != NULL)
		kfree(pl->name);

	pl->name = newname;

	if (pl->_state == NJB_PL_UNCHANGED)
		pl->_state = NJB_PL_CHNAME;

	return 0;
}

static playlist_track_t *playlist_track_new(u_int32_t trackid)
{
	playlist_track_t *track;

	track =
	    (playlist_track_t *) kmalloc(sizeof(playlist_track_t),
					 GFP_KERNEL);
	if (track == NULL) {
		return NULL;
	}

	memset(track, 0, sizeof(playlist_track_t));
	track->trackid = trackid;

	return track;
}

static void playlist_track_destroy(playlist_track_t * track)
{
	kfree(track);
}

EXPORT_SYMBOL(playlist_set_name);
EXPORT_SYMBOL(playlist_deltrack);
EXPORT_SYMBOL(playlist_unpack);
