/*
 * Nomad Jukebox filesystem for Linux.
 *
 * Copyright (C) 2001 Jrg Prante <joergprante@gmx.de>
 *
 * based on ftpfs by Florin Malita <fmalita@yahoo.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


/**
 *  Processing the directives from the VFS and delegate them to the
 *  NJB USB subsystem
 */


#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/smp_lock.h>
#include <linux/fs.h>
#include <linux/usb.h>
#include <linux/time.h>
#include <linux/ctype.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/errno.h>

#include "njbfs.h"
#include "nomad_usb.h"
#include "track.h"


/* njb_usb */
extern int njb_usb_capture(struct nomad_usb_data *njb);
extern int njb_usb_release(struct nomad_usb_data *njb);
extern int njb_usb_get_first_track_tag_header(struct nomad_usb_data *njb,
					      njbttaghdr_t * header);
extern int njb_usb_get_next_track_tag_header(struct nomad_usb_data *njb,
					     njbttaghdr_t * header);
extern track_t *njb_usb_get_track_tag(struct nomad_usb_data *njb,
				      njbttaghdr_t * header);
extern int njb_usb_connect(struct nomad_usb_data *njb);
extern int njb_usb_disconnect(struct nomad_usb_data *njb);
extern int njb_usb_open(struct nomad_usb_data *njb,
			struct njbfs_fattr *fattr);
extern int njb_usb_close(struct nomad_usb_data *njb,
			 struct njbfs_fattr *fattr);
extern int njb_usb_write(struct nomad_usb_data *njb, char *name,
			 u_int32_t offset, u_int32_t count, void *buffer);
extern int njb_usb_read(struct nomad_usb_data *njb, char *name,
			u_int32_t offset, u_int32_t count, void *buffer);
extern int njb_usb_rename(struct nomad_usb_data *njb,
			  struct njbfs_fattr *fattr);
extern int njb_usb_delete(struct nomad_usb_data *njb, char *file);
extern int njb_usb_create(struct nomad_usb_data *njb, char *file);

/* forward decl */
static int njbfs_parse_filename(struct njbfs_fattr *fattr,
				const char *file);
unsigned char *k_parse_attribute(unsigned char *buf, unsigned char *key);



/**
 *   Mount file system
 */
int njbfs_connect(struct njbfs_sb_info *info)
{
	return njb_usb_connect(info->njb);
}


/**
 *   Unmount file system
 */
int njbfs_disconnect(struct njbfs_sb_info *info)
{
	return njb_usb_disconnect(info->njb);
}


/**
 *  Initialize root directory
 */
void njbfs_init_root_dirent(struct njbfs_sb_info *server,
			    struct njbfs_fattr *fattr)
{
	memset(fattr, 0, sizeof(*fattr));
	fattr->f_nlink = 1;
	fattr->f_uid = server->mnt.uid;
	fattr->f_gid = server->mnt.gid;
	fattr->f_blksize = NJBFS_BLOCKSIZE;

	fattr->f_ino = 2;
	fattr->f_mtime = CURRENT_TIME;
	fattr->f_mode =
	    S_IRUSR | S_IRGRP | S_IROTH | S_IXUSR | S_IXGRP | S_IXOTH |
	    S_IFDIR | server->mnt.dir_mode;
	fattr->f_size = NJBFS_BLOCKSIZE;
	fattr->f_blocks = 0;
}


int njbfs_parse_mode(char *mode)
{
	return ((mode[0] - '0') << 6) + ((mode[1] - '0') << 3) + (mode[2] -
								  '0');
}

/**
 *  Parse directory
  */
int njbfs_parse_options(struct njbfs_sb_info *info, void *opts)
{
	char *p;
	int optcnt = 0;

	strcpy(info->virtual_dir, "tracks");

	if (!opts)
		return 0;
	p = strtok(opts, ",");
	for (; p; p = strtok(NULL, ",")) {
		if (strncmp(p, "dir=", 5) == 0) {
			if (strlen(p + 5) > NJBFS_MAXVIRTUALDIRLEN) {
				err("ugly opts, go away!");
				return -EINVAL;
			}
			strcpy(info->virtual_dir, p + 5);
			optcnt++;
		}
	}
	return optcnt;
}


int njbfs_get_fname(char *s, char *d)
{
	int i, cnt;
	int in_space = 1;

	for (i = 0, cnt = 9; (i < strlen(s)) && (cnt > 0); i++) {
		if (in_space) {
			if (s[i] != ' ') {
				in_space = 0;
				cnt--;
			}
		} else {
			if (s[i] == ' ')
				in_space = 1;
		}
	}
	if (i >= strlen(s)) {
		d[0] = 0;
		return -EINVAL;
	}
	strcpy(d, &(s[i - 1]));
	d[strlen(d) - 2] = 0;
	return 0;
}


u_int32_t get_track_size(track_frame_t * frame)
{
	if (frame->datasz == 1) {
		char tval = ((char *) frame->data)[0];
		return (u_int32_t) tval;
	} else if (frame->datasz == 2) {
		u_int16_t tval;
		memcpy(&tval, frame->data, 2);
		return (u_int32_t) tval;
	} else if (frame->datasz == 4) {
		u_int32_t tval;
		memcpy(&tval, frame->data, 4);
		return tval;
	}

	return 0;
}


/**
 *  Get directory node from a track structure
 */
int njbfs_get_dirlist_node_from_track(struct njbfs_dirlist_node *p,
				      track_t * track)
{
	char buf[NJBFS_MAXPATHLEN];
	track_frame_t *frame;
	char *artist = NULL;
	char *title = NULL;
	char *codec = NULL;
	char *fname = NULL;
	char *ch;
	u_int32_t filesize = 0;
	int status;

	track->cur = track->first;
	frame = track->cur;

	while (frame != NULL) {
		if (!strcmp(FR_ARTIST, frame->label))
			artist = (char *) frame->data;
		else if (!strcmp(FR_TITLE, frame->label))
			title = (char *) frame->data;
		else if (!strcmp(FR_CODEC, frame->label))
			codec = (char *) frame->data;
		else if (!strcmp(FR_SIZE, frame->label))
			filesize = get_track_size(frame);
		else if (!strcmp(FR_FNAME, frame->label))
			fname = (char *) frame->data;

		frame = track_getframe(track);
	}

	/* file name */
	if (fname) {
		strcpy(buf, fname);
	}
	if ((!artist) || (!title) || (!codec))
		return -EINVAL;

	/* codec to lower case (MP3 ->mp3) */
	ch = codec;
	while ((*ch = tolower(*ch)) != 0)
		ch++;

	/*
	   Replace characters like '?', '*' or '/' in
	   artist and title with placeholder
	 */
	ch = artist;
	while (*ch) {
		switch (*ch) {
		case '/':
			*ch = '_';
		case '*':
			*ch = '_';
		case '?':
			*ch = '_';
		}
		ch++;
	}
	ch = title;
	while (*ch) {
		switch (*ch) {
		case '/':
			*ch = '_';
		case '*':
			*ch = '_';
		case '?':
			*ch = '_';
		}
		ch++;
	}

	status = sprintf(buf, "%s - %s.%s", artist, title, codec);

	p->entry.name = (char *) kmalloc(strlen(buf) + 1, GFP_KERNEL);
	if (!p->entry.name) {
		err("kmalloc error");
		return -ENOMEM;
	}
	strcpy(p->entry.name, buf);

	/* attributes */
	p->entry.size = filesize;
	p->entry.blocksize = NJBFS_BLOCKSIZE;
	p->entry.blocks =
	    (p->entry.size +
	     (NJBFS_BLOCKSIZE - 1)) >> (NJBFS_BLOCKSIZE_BITS - 1);

	p->entry.mode |=
	    S_IFREG | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH |
	    S_IWOTH;
	p->entry.atime = CURRENT_TIME;

	p->entry.fileid = track->header.trackid;
	p->entry.track = track;

	return 0;
}



/**
 *  Get directory entries from NJB
 */
int njbfs_loaddir(struct njbfs_sb_info *info, char *name,
		  struct njbfs_directory *dir)
{
	struct njbfs_dirlist_node *p;

	struct nomad_usb_data *njb = info->njb;
	njbttaghdr_t header;
	track_t *track;
	int status;

	if (strlen(info->mnt.root) + strlen(name) > NJBFS_MAXPATHLEN) {
		err("loaddir: path too long");
		return -EINVAL;
	}

	/* capture njb */
	status = njb_usb_capture(njb);
	if (status < 0) {
		err("loaddir: capture failed, status=%d", status);
		return -EIO;
	}

	status = njb_usb_get_first_track_tag_header(njb, &header);

	while (status != -ENOENT) {
		track = njb_usb_get_track_tag(njb, &header);
		if (track == NULL)
			break;

		p =
		    (struct njbfs_dirlist_node *)
		    kmalloc(sizeof(struct njbfs_dirlist_node), GFP_KERNEL);
		if (!p) {
			err("loaddir: kmalloc error");
			return -ENOMEM;
		}
		memset(p, 0, sizeof(struct njbfs_dirlist_node));

		status = njbfs_get_dirlist_node_from_track(p, track);
		if (status < 0) {
			err
			    ("loaddir: unable to get dirlist node from track");
			kfree(p);
			return -EIO;
		}

		p->prev = NULL;
		p->next = dir->head;
		dir->head = p;

		status = njb_usb_get_next_track_tag_header(njb, &header);
	}

	/* release njb */
	status = njb_usb_release(njb);
	if (status < 0) {
		err("loaddir: release failed, status=%d", status);
		return -EIO;
	}
	return 0;

}



int njbfs_get_name(struct dentry *d, char *name)
{
	int len = 0;
	struct dentry *p;

	for (p = d; p != p->d_parent; p = p->d_parent)
		len += p->d_name.len + 1;

	if (len > NJBFS_MAXPATHLEN)
		return -EINVAL;
	if (len == 0) {
		name[0] = '/';
		name[1] = 0;
		return 0;
	}

	name[len] = 0;
	for (p = d; p != p->d_parent; p = p->d_parent) {
		len -= p->d_name.len;
		strncpy(&(name[len]), p->d_name.name, p->d_name.len);
		len--;
		name[len] = '/';
	}

	return 0;
}


inline void njbfs_lock(struct njbfs_sb_info *info)
{
	down(&(info->sem));
}

inline void njbfs_unlock(struct njbfs_sb_info *info)
{
	up(&(info->sem));
}


/**
 *  Get attributes of a dentry
 */

int njbfs_get_attr(struct dentry *dentry, struct njbfs_fattr *fattr,
		   struct njbfs_sb_info *info)
{
	struct njbfs_directory *dir;
	char buf[NJBFS_MAX_LINE];
	struct njbfs_dirlist_node *file;
	int status;

	/* load parent dir cache */
	njbfs_get_name(dentry->d_parent, buf);
	njbfs_lock(info);
	status = njbfs_cache_get(info, buf, &dir);
	njbfs_unlock(info);
	if (status < 0) {
		err("njbfs_get_attr: cache get failed");
		return status;
	}

	/* scan dir cache */
	for (file = dir->head; file != NULL; file = file->next) {
		/* look for artist, title, album, genre */
		if (strcmp(dentry->d_name.name, file->entry.name) == 0)
			break;
	}

	if (!file) {
		fattr->f_mode =
		    S_IFREG | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP |
		    S_IROTH | S_IWOTH;
		fattr->f_size = NJBFS_BLOCKSIZE;
		fattr->f_blksize = NJBFS_BLOCKSIZE;
		fattr->f_blocks = 0;
		fattr->f_atime = CURRENT_TIME;
		fattr->f_mtime = CURRENT_TIME;
		fattr->f_uid = info->mnt.uid;
		fattr->f_gid = info->mnt.gid;

		/* NJB attributes */
		fattr->fileid = 0;
		fattr->track = NULL;
		fattr->size = 0;

		status = njbfs_parse_filename(fattr, dentry->d_name.name);

		return status;
	}

	fattr->f_mode = file->entry.mode;
	if (file->entry.mode & S_IFDIR)
		fattr->f_mode |= info->mnt.dir_mode;
	else
		fattr->f_mode |= info->mnt.file_mode;

	fattr->f_size = file->entry.size;
	fattr->f_blksize = file->entry.blocksize;
	fattr->f_blocks = file->entry.blocks;
	fattr->f_nlink = file->entry.nlink;
	fattr->f_atime = file->entry.atime;

	/*Any way to get the real mtime, ctime? */
	fattr->f_mtime = file->entry.atime;
	fattr->f_ctime = file->entry.atime;
	fattr->f_uid = info->mnt.uid;
	fattr->f_gid = info->mnt.gid;

	/* get the NJB attributes */
	fattr->fileid = file->entry.fileid;
	fattr->track = NULL;
	fattr->size = 0;

	status = njbfs_parse_filename(fattr, dentry->d_name.name);

	return status;

}


void njbfs_free_attr(struct njbfs_fattr *fattr)
{
	if (!fattr) {
		return;
	}

	if (fattr->codec)
		kfree(fattr->codec);
	fattr->codec = NULL;

	if (fattr->title)
		kfree(fattr->title);
	fattr->title = NULL;

	if (fattr->artist)
		kfree(fattr->artist);
	fattr->artist = NULL;

	if (fattr->album)
		kfree(fattr->album);
	fattr->album = NULL;

	if (fattr->genre)
		kfree(fattr->genre);
	fattr->genre = NULL;

	if (fattr->track)
		kfree(fattr->track);
	fattr->track = NULL;

}


/**
 *  Open a file
 */
int njbfs_open(struct inode *inode, struct file *f)
{
	struct super_block *sb = inode->i_sb;
	struct njbfs_sb_info *info =
	    (struct njbfs_sb_info *) sb->u.generic_sbp;
	struct dentry *dentry = f->f_dentry;
	struct njbfs_fattr fattr;
	int status;

	if ((status = njbfs_get_attr(dentry, &fattr, info)) < 0) {
		info("open: file not found");
		njbfs_free_attr(&fattr);
		return status;
	}

	dbg("open: opening file, file id=%d", fattr.fileid);

	status = njb_usb_open(info->njb, &fattr);

	njbfs_free_attr(&fattr);

	return status;

}

/**
 *  Close a file
 */
int njbfs_close(struct inode *inode, struct file *f)
{
	struct super_block *sb = inode->i_sb;
	struct njbfs_sb_info *info =
	    (struct njbfs_sb_info *) sb->u.generic_sbp;
	struct dentry *dentry = f->f_dentry;
	struct njbfs_fattr fattr;
	int status;

	if ((status = njbfs_get_attr(dentry, &fattr, info)) < 0) {
		err("njbfs_close: file not found, status=%d", status);
		goto out;
	}

	dbg("close: closing file, file id=%d", fattr.fileid);

	status = njb_usb_close(info->njb, &fattr);

      out:
	njbfs_free_attr(&fattr);
	return status;

}


/**
 *  Read from a file. This is a function called from the VFS
 *
 *  @param dentry  directory entry to read from
 *  @param offset  position in the file
 *  @param count  how many bytes we can read
 *  @param buffer the buffer to write the data
 *  @return the number of bytes which could be read
 */
int njbfs_read(struct dentry *dentry, unsigned long offset,
	       unsigned long count, char *buffer)
{
	struct inode *inode = dentry->d_inode;
	struct super_block *sb = inode->i_sb;
	struct njbfs_sb_info *info =
	    (struct njbfs_sb_info *) sb->u.generic_sbp;
	char buf[NJBFS_MAXPATHLEN], name[NJBFS_MAXPATHLEN];
	int bytes;

	njbfs_get_name(dentry, buf);

	sprintf(name, "%s%s", info->mnt.root, buf);

	if (strlen(info->mnt.root) + strlen(buf) > NJBFS_MAXPATHLEN) {
		info("read: path too long");
		return -EINVAL;
	}

	/* transfer */
	njbfs_lock(info);
	if (
	    (bytes =
	     njb_usb_read(info->njb, name, (u_int32_t) offset,
			  (u_int32_t) count, (void *) buffer)) < 0) {
		err("read: couldn't read data from %s", name);
		njbfs_unlock(info);
		return bytes;
	}
	njbfs_unlock(info);

	return bytes;
}

/**
 *  Write to a file
 *
 *  @param dentry  directory entry to read from
 *  @param offset  position in the file
 *  @param count  how many bytes we can read
 *  @param buffer the buffer to write the data
 *  @return the number of bytes which could be written
 */
int njbfs_write(struct dentry *dentry, unsigned long offset,
		unsigned long count, char *buffer)
{
	struct inode *inode = dentry->d_inode;
	struct super_block *sb = inode->i_sb;
	struct njbfs_sb_info *info =
	    (struct njbfs_sb_info *) sb->u.generic_sbp;
	char buf[NJBFS_MAXPATHLEN], name[NJBFS_MAXPATHLEN];
	int bytes;

	njbfs_get_name(dentry, buf);

	sprintf(name, "%s%s", info->mnt.root, buf);

	if (strlen(buf) + strlen(info->mnt.root) > NJBFS_MAXPATHLEN) {
		err("write: path too long");
		return -EINVAL;
	}

	/* transfer */

	njbfs_lock(info);
	if (
	    (bytes =
	     njb_usb_write(info->njb, name, (u_int32_t) offset,
			   (u_int32_t) count, (void *) buffer)) < 0) {
		err("write: couldn't write data to %s", name);
		njbfs_unlock(info);
		return bytes;
	}
	njbfs_unlock(info);

	/* unload dir cache */
/*
    njbfs_cache_invalidate(dentry->d_parent);
    info("write: offset=%ld, count=%ld, bytes=%d", offset, count, bytes);
*/
	return bytes;
}

/**
 *  Create a directory
 */
int njbfs_proc_mkdir(struct njbfs_sb_info *info, char *dir)
{
	char buf[NJBFS_MAXPATHLEN + 5];
	int error = 0;

	if (strlen(info->mnt.root) + strlen(dir) > NJBFS_MAXPATHLEN) {
		info("path too long");
		return -EINVAL;
	}
	sprintf(buf, "%s%s", info->mnt.root, dir);

	njbfs_lock(info);
	/* not much to do? */
	njbfs_unlock(info);

	return error;
}

/**
 *  Remove a directory
 */
int njbfs_proc_rmdir(struct njbfs_sb_info *info, char *dir)
{
	char buf[NJBFS_MAXPATHLEN + 5];
	int error = 0;

	if (strlen(info->mnt.root) + strlen(dir) > NJBFS_MAXPATHLEN) {
		err("rmdir: path too long");
		return -EINVAL;
	}
	sprintf(buf, "%s%s", info->mnt.root, dir);

	njbfs_lock(info);
	/* not much to do? */
	njbfs_unlock(info);

	return error;
}


/**
 * Rename file or directory
 */
int
njbfs_proc_rename(struct njbfs_sb_info *info, struct inode *old_inode,
		  struct dentry *old_dentry, struct inode *new_inode,
		  struct dentry *new_dentry)
{
	char old_buf[NJBFS_MAXPATHLEN + 6], new_buf[NJBFS_MAXPATHLEN + 6];
	struct njbfs_fattr old_fattr, new_fattr;
	int status;

	njbfs_get_name(old_dentry, old_buf);
	njbfs_get_name(new_dentry, new_buf);

	if ((status = njbfs_get_attr(old_dentry, &old_fattr, info)) < 0) {
		err("rename: file '%s' not found", old_buf);
		njbfs_free_attr(&old_fattr);
		return status;
	}

	njbfs_lock(info);

	/* copy fattr */
	new_fattr.f_mode = old_fattr.f_mode;
	new_fattr.f_size = old_fattr.f_mode;
	new_fattr.f_blksize = old_fattr.f_blksize;
	new_fattr.f_blocks = old_fattr.f_blocks;
	new_fattr.f_atime = CURRENT_TIME;
	new_fattr.f_mtime = CURRENT_TIME;
	new_fattr.f_uid = old_fattr.f_uid;
	new_fattr.f_gid = old_fattr.f_gid;
	new_fattr.fileid = old_fattr.fileid;
	new_fattr.track = NULL;
	new_fattr.size = old_fattr.size;
	/* default attributes */
	new_fattr.codec = old_fattr.codec;
	new_fattr.title = old_fattr.title;
	new_fattr.artist = old_fattr.artist;
	new_fattr.album = old_fattr.album;
	new_fattr.genre = old_fattr.genre;
	/* parse new attributes from file name */
	status = njbfs_parse_filename(&new_fattr, new_buf);

	/* rename */

	if ((status = njb_usb_rename(info->njb, &new_fattr)) < 0) {
		err("rename: couldn't rename!");
	}
	njbfs_unlock(info);

	return status;
}

/**
 *  Delete a file
 */
int njbfs_proc_unlink(struct njbfs_sb_info *info, char *file)
{
	char buf[NJBFS_MAXPATHLEN + 6];
	int status;

	if (strlen(info->mnt.root) + strlen(file) > NJBFS_MAXPATHLEN) {
		err("unlink: path too long!");
		return -EINVAL;
	}
	sprintf(buf, "%s%s", info->mnt.root, file);

	njbfs_lock(info);
	if ((status = njb_usb_delete(info->njb, buf)) < 0) {
		err("unlink: couldn't delete file!");
		status = -EIO;
	}
	njbfs_unlock(info);
	return status;
}


/**
 *  Create a new file
 *  A created file is being reopened before written to.
 *  We just parse the file name and that's it.
 */
int njbfs_proc_create(struct njbfs_sb_info *info, char *file)
{
	char buf[NJBFS_MAXPATHLEN + 6];
	int status;

	if (strlen(info->mnt.root) + strlen(file) > NJBFS_MAXPATHLEN) {
		err("create: path too long!");
		return -EINVAL;
	}
	sprintf(buf, "%s%s", info->mnt.root, file);

	njbfs_lock(info);


	if ((status = njb_usb_create(info->njb, buf)) < 0) {
		err("create: couldn't create file!");
		status = -EINVAL;
	}
	njbfs_unlock(info);

	return status;
}



/**
 * Parse a file name for title, artist, codec, album, genre, size
 */
static int njbfs_parse_filename(struct njbfs_fattr *fattr,
				const char *file)
{
	char buf[NJBFS_MAXPATHLEN];
	char buf2[NJBFS_MAXPATHLEN];
	char *p, *q;

	strcpy(buf, file);

	/* parse size */
	if ((p = strstr(buf, ",size=")) != NULL) {
		strcpy(buf2, p + 6);
		sscanf(buf2, "%u", &fattr->size);
		*p = '\0';
	} else {
		p = k_parse_attribute(buf, "size");
		if (p) {
			sscanf(p, "%u", &fattr->size);
			kfree(p);
		} else
			fattr->size = 0;
	}

	/* parse codec */
	if ((strstr(buf, ".mp3") != NULL) || (strstr(buf, ".MP3") != NULL)) {
		fattr->codec =
		    (char *) kmalloc(strlen(NJB_CODEC_MP3) + 1,
				     GFP_KERNEL);
		strcpy(fattr->codec, NJB_CODEC_MP3);
	} else {
		p = k_parse_attribute(buf, "codec");
		if (p && (!strcmp(p, "mp3") || !strcmp(p, "MP3"))) {
			fattr->codec =
			    (char *) kmalloc(strlen(NJB_CODEC_MP3) + 1,
					     GFP_KERNEL);
			strcpy(fattr->codec, NJB_CODEC_MP3);
			kfree(p);
		} else {	/* default is MP3 */
			fattr->codec =
			    (char *) kmalloc(strlen(NJB_CODEC_MP3) + 1,
					     GFP_KERNEL);
			strcpy(fattr->codec, NJB_CODEC_MP3);
		}
	}

	/* parse title */
	if ((p = strstr(buf, " - ")) != NULL) {
		/* cut '.mp3' suffix */
		q = strrchr(buf, '.');
		if (q) {
			*q = '\0';
		}
		fattr->title =
		    (char *) kmalloc(strlen(p + 3) + 1, GFP_KERNEL);
		strcpy(fattr->title, p + 3);
	} else {
		fattr->title = k_parse_attribute(buf, "title");
		if (!fattr->title) {
			fattr->title =
			    (char *) kmalloc(strlen("<title unknown>") + 1,
					     GFP_KERNEL);
			strcpy(fattr->title, "<title unknown>");
		}
	}

	/* parse artist */
	if (p) {
		*p = '\0';
	}
	fattr->artist = k_parse_attribute(buf, "artist");
	if (!fattr->artist) {
		fattr->artist =
		    (char *) kmalloc(strlen(buf) + 1, GFP_KERNEL);
		strcpy(fattr->artist, buf);
	}

	/* parse album */
	fattr->album = k_parse_attribute(buf, "album");
	if (!fattr->album) {
		fattr->album =
		    (char *) kmalloc(strlen("<unknown>") + 1, GFP_KERNEL);
		strcpy(fattr->album, "<unknown>");
	}

	/* parse genre */
	fattr->genre = k_parse_attribute(buf, "genre");
	if (!fattr->genre) {
		fattr->genre =
		    (char *) kmalloc(strlen("<unknown>") + 1, GFP_KERNEL);
		strcpy(fattr->genre, "<unknown>");
	}

	/* parse length in milliseconds */
	p = k_parse_attribute(buf, "length");
	if (p) {
		sscanf(p, "%u", &fattr->length);
		kfree(p);
	} else {
		fattr->length = 0;
	}

	/* parse track num */
	p = k_parse_attribute(buf, "tracknum");
	if (p) {
		sscanf(p, "%u", &fattr->tracknum);
		kfree(p);
	} else {
		fattr->tracknum = 0;
	}

	return 0;
}

/**
 +  Helper function
 *  Looking for a key in a buffer, and, extract value from buffer into the parameter 'value'.
 *  buf will be modified and does not contain the attribute any more.
 */

unsigned char *k_parse_attribute(unsigned char *buf, unsigned char *key)
{
	unsigned char *p, *right = NULL, *left = NULL, *value;
	int len = 0;

	p = strstr(buf, key);

	if (!p)
		return NULL;

	left = p + strlen(key);

	while (*left) {
		if (*left == '=')
			break;
		left++;
	}
	if (!(*left))
		return NULL;

	left++;
	while (*left) {
		if (*left == '{')
			break;
		left++;
	}

	if (left) {
		left++;
		right = left;
		while (*right) {
			if (*right == '}')
				break;
			right++;
			len++;
		}
	}

	value = (unsigned char *) kmalloc(len + 1, GFP_KERNEL);
	memset(value, 0, len + 1);
	memcpy(value, left, len);

	if (*left == '\0')
		*p = '\0';
	else {
		if (*right)
			right++;
		while (*right)
			*p++ = *right++;
		*p = '\0';
	}

	return value;

}

/*
 *  2.4.10- kernel compatibility hack for missing 'sscanf' symbol
 *  The following lines are copied from 2.4.10+ Linux kernel lib/vsprintf.c
 *  They can be removed if people get used to kernel 2.4.10+
 */



/**
 * simple_strtoull - convert a string to an unsigned long long
 * @cp: The start of the string
 * @endp: A pointer to the end of the parsed string will be placed here
 * @base: The number base to use
 */
unsigned long long simple_strtoull(const char *cp, char **endp,
				   unsigned int base)
{
	unsigned long long result = 0, value;

	if (!base) {
		base = 10;
		if (*cp == '0') {
			base = 8;
			cp++;
			if ((*cp == 'x') && isxdigit(cp[1])) {
				cp++;
				base = 16;
			}
		}
	}
	while (isxdigit(*cp)
	       && (value = isdigit(*cp) ? *cp - '0' : (islower(*cp)
						       ? toupper(*cp) :
						       *cp) - 'A' + 10) <
	       base) {
		result = result * base + value;
		cp++;
	}
	if (endp)
		*endp = (char *) cp;
	return result;
}

/**
 * simple_strtoll - convert a string to a signed long long
 * @cp: The start of the string
 * @endp: A pointer to the end of the parsed string will be placed here
 * @base: The number base to use
 */
long long simple_strtoll(const char *cp, char **endp, unsigned int base)
{
	if (*cp == '-')
		return -simple_strtoull(cp + 1, endp, base);
	return simple_strtoull(cp, endp, base);
}


static int skip_atoi(const char **s)
{
	int i = 0;

	while (isdigit(**s))
		i = i * 10 + *((*s)++) - '0';
	return i;
}

/**
 *
 * vsscanf - Unformat a buffer into a list of arguments
 * @buf:	input buffer
 * @fmt:	format of buffer
 * @args:	arguments
 */
int vsscanf(const char *buf, const char *fmt, va_list args)
{
	const char *str = buf;
	char *next;
	int num = 0;
	int qualifier;
	int base;
	int field_width = -1;
	int is_sign = 0;

	while (*fmt && *str) {
		/* skip any white space in format */
		/* white space in format matchs any amount of
		 * white space, including none, in the input.
		 */
		if (isspace(*fmt)) {
			while (isspace(*fmt))
				++fmt;
			while (isspace(*str))
				++str;
		}

		/* anything that is not a conversion must match exactly */
		if (*fmt != '%' && *fmt) {
			if (*fmt++ != *str++)
				break;
			continue;
		}

		if (!*fmt)
			break;
		++fmt;

		/* skip this conversion.
		 * advance both strings to next white space
		 */
		if (*fmt == '*') {
			while (!isspace(*fmt) && *fmt)
				fmt++;
			while (!isspace(*str) && *str)
				str++;
			continue;
		}

		/* get field width */
		if (isdigit(*fmt))
			field_width = skip_atoi(&fmt);

		/* get conversion qualifier */
		qualifier = -1;
		if (*fmt == 'h' || *fmt == 'l' || *fmt == 'L'
		    || *fmt == 'Z') {
			qualifier = *fmt;
			fmt++;
		}
		base = 10;
		is_sign = 0;

		if (!*fmt || !*str)
			break;

		switch (*fmt++) {
		case 'c':
			{
				char *s = (char *) va_arg(args, char *);
				if (field_width == -1)
					field_width = 1;
				do {
					*s++ = *str++;
				} while (field_width-- > 0 && *str);
				num++;
			}
			continue;
		case 's':
			{
				char *s = (char *) va_arg(args, char *);
				if (field_width == -1)
					field_width = INT_MAX;
				/* first, skip leading white space in buffer */
				while (isspace(*str))
					str++;

				/* now copy until next white space */
				while (*str && !isspace(*str)
				       && field_width--) {
					*s++ = *str++;
				}
				*s = '\0';
				num++;
			}
			continue;
		case 'n':
			/* return number of characters read so far */
			{
				int *i = (int *) va_arg(args, int *);
				*i = str - buf;
			}
			continue;
		case 'o':
			base = 8;
			break;
		case 'x':
		case 'X':
			base = 16;
			break;
		case 'd':
		case 'i':
			is_sign = 1;
		case 'u':
			break;
		case '%':
			/* looking for '%' in str */
			if (*str++ != '%')
				return num;
			continue;
		default:
			/* invalid format; stop here */
			return num;
		}

		/* have some sort of integer conversion.
		 * first, skip white space in buffer.
		 */
		while (isspace(*str))
			str++;

		if (!*str || !isdigit(*str))
			break;

		switch (qualifier) {
		case 'h':
			if (is_sign) {
				short *s = (short *) va_arg(args, short *);
				*s =
				    (short) simple_strtol(str, &next,
							  base);
			} else {
				unsigned short *s =
				    (unsigned short *) va_arg(args,
							      unsigned
							      short *);
				*s =
				    (unsigned short) simple_strtoul(str,
								    &next,
								    base);
			}
			break;
		case 'l':
			if (is_sign) {
				long *l = (long *) va_arg(args, long *);
				*l = simple_strtol(str, &next, base);
			} else {
				unsigned long *l =
				    (unsigned long *) va_arg(args,
							     unsigned long
							     *);
				*l = simple_strtoul(str, &next, base);
			}
			break;
		case 'L':
			if (is_sign) {
				long long *l =
				    (long long *) va_arg(args,
							 long long *);
				*l = simple_strtoll(str, &next, base);
			} else {
				unsigned long long *l =
				    (unsigned long long *) va_arg(args,
								  unsigned
								  long long
								  *);
				*l = simple_strtoull(str, &next, base);
			}
			break;
		case 'Z':
			{
				size_t *s =
				    (size_t *) va_arg(args, size_t *);
				*s =
				    (size_t) simple_strtoul(str, &next,
							    base);
			}
			break;
		default:
			if (is_sign) {
				int *i = (int *) va_arg(args, int *);
				*i = (int) simple_strtol(str, &next, base);
			} else {
				unsigned int *i =
				    (unsigned int *) va_arg(args,
							    unsigned int
							    *);
				*i =
				    (unsigned int) simple_strtoul(str,
								  &next,
								  base);
			}
			break;
		}
		num++;

		if (!next)
			break;
		str = next;
	}
	return num;
}

/**
 * sscanf - Unformat a buffer into a list of arguments
 * @buf:	input buffer
 * @fmt:	formatting of buffer
 * @...:	resulting arguments
 */
int sscanf(const char *buf, const char *fmt, ...)
{
	va_list args;
	int i;

	va_start(args, fmt);
	i = vsscanf(buf, fmt, args);
	va_end(args);
	return i;
}

/* end of 'sscanf' kernel compatibility hack */

EXPORT_SYMBOL(njbfs_open);
EXPORT_SYMBOL(njbfs_close);
EXPORT_SYMBOL(njbfs_connect);
EXPORT_SYMBOL(njbfs_disconnect);
